<?php

namespace Ankou;

use Exception;
use FML\ManiaLink;
use FML\Controls\Frame;
use FML\Controls\Label;
use FML\Controls\Quad;
use FML\Controls\Labels\Label_Text;
use FML\Controls\Quads\Quad_Icons64x64_1;
use FML\Script\Features\ActionTrigger;
use FML\Script\Features\MapInfo;
use ManiaControl\Admin\AuthenticationManager;
use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\Structures\Common\StatusCallbackStructure;
use ManiaControl\Callbacks\Structures\ShootMania\OnShootStructure;
use ManiaControl\Callbacks\Structures\TrackMania\OnScoresStructure;
use ManiaControl\Commands\CommandListener;
use ManiaControl\Communication\CommunicationAnswer;
use ManiaControl\Communication\CommunicationListener;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Plugins\Plugin;
use ManiaControl\ManiaControl;
use \ManiaControl\Logger;
use ManiaControl\Settings\Setting;
use ManiaControl\Settings\SettingManager;
use ManiaControl\Configurator\GameModeSettings;
use ManiaControl\Utils\Formatter;
use Maniaplanet\DedicatedServer\InvalidArgumentException;
use ManiaControl\Callbacks\TimerListener; // for pause
use ManiaControl\Files\AsyncHttpRequest;
use ManiaControl\Maps\Map;

class YoxServer
{
	public $id;
	public $login;
	public $name;



	public static function unserialize($json)
	{
		$server = new YoxServer();

		if (isset($json->{"id"}))
			$server->id = $json->{"id"};
		if (isset($json->{"name"}))
			$server->name = $json->{"name"};
		if (isset($json->{"login"}))
			$server->name = $json->{"login"};


		return $server;
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxCup
{
	public $id;
	public $name;
	public $description;
	public $dateCrea;
	public $dateCup;
	public $dateRegisterLimit;
	public $cupMode;
	public $gameMode;
	public $registerMode;
	public $nPlayerByTeam;
	public $teams;
	public $players;
	public $nMapPerMatch;
	public $nRoundPerMatch;
	public $matchs;


	public static function unserialize($json)
	{
		$cup = new YoxCup();

		if (isset($json->{"id"}))
			$cup->id = $json->{"id"};
		if (isset($json->{"name"}))
			$cup->name = $json->{"name"};
		if (isset($json->{"description"}))
			$cup->description = $json->{"description"};
		if (isset($json->{"dateCrea"}))
			$cup->dateCrea = $json->{"dateCrea"};
		if (isset($json->{"dateCup"}))
			$cup->dateCup = $json->{"dateCup"};
		if (isset($json->{"dateRegisterLimit"}))
			$cup->dateRegisterLimit = $json->{"dateRegisterLimit"};
		if (isset($json->{"cupMode"}))
			$cup->cupMode = $json->{"cupMode"}->{"name"};
		if (isset($json->{"cupGameMode"}))
			$cup->gameMode = $json->{"cupGameMode"}->{"name"};
		if (isset($json->{"registerMode"}->{"name"}))
			$cup->registerMode = $json->{"registerMode"}->{"name"};
		if (isset($json->{"nPlayerByTeam"}))
			$cup->nPlayerByTeam = $json->{"nPlayerByTeam"};
		if (isset($json->{"nMapPerMatch"}))
			$cup->nMapPerMatch = $json->{"nMapPerMatch"};
		if (isset($json->{"nRoundPerMatch"}))
			$cup->nRoundPerMatch = $json->{"nRoundPerMatch"};

		if (isset($json->{"teams"}))
			$cup->teams = YoxTeam::unserializeArray($json->{"teams"});
		if (isset($json->{"matchs"}))
			$cup->matchs = YoxMatch::unserializeArray($json->{"matchs"});
		if (isset($json->{"players"}))
			$cup->players = YoxPlayer::unserializeArray($json->{"players"});

		return $cup;
	}

	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxTeam
{
	public $id;
	public $name;
	public $avatar;
	public $admin;
	public $players;

	public static function unserialize($json)
	{
		$team = new YoxTeam();


		if (isset($json->{"id"}))
			$team->id = $json->{"id"};
		if (isset($json->{"name"}))
			$team->name = $json->{"name"};
		if (isset($json->{"avatar"}))
			$team->avatar = $json->{"avatar"};
		if (isset($json->{"players"}))
			$team->players = YoxPlayer::unserializeArray($json->{"players"});

		return $team;
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxPlayer
{
	public $id;
	public $login;
	public $username;
	public $level;
	public $teams;

	public static function unserialize($json)
	{
		$player = new YoxPlayer();

		if (isset($json->{"id"}))
			$player->id = $json->{"id"};
		if (isset($json->{"login"}))
			$player->login = $json->{"login"};
		if (isset($json->{"username"}))
			$player->username = $json->{"username"};
		if (isset($json->{"level"}))
			$player->level = $json->{"level"};
		if (isset($json->{"teams"}))
			$player->teams = YoxTeam::unserializeArray($json->{"teams"});


		return $player;
	}

	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxMap
{
	public static function unserialize($json)
	{
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxMapPack
{
	public static function unserialize($json)
	{
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxMatch
{
	public $id;
	public $cup;
	public $nMatch;
	public $dateMatch;
	public $teamA;
	public $teamB;
	public $mapPack;
	public $server;


	public static function unserialize($json)
	{
		$match = new YoxMatch();

		if (isset($json->{"id"}))
			$match->id = $json->{"id"};
		if (isset($json->{"nMatch"}))
			$match->nMatch = $json->{"nMatch"};
		if (isset($json->{"cup"}))
			$match->cup = YoxCup::unserialize($json->{"cup"});
		if (isset($json->{"dateMatch"}))
			$match->dateMatch = $json->{"dateMatch"};
		if (isset($json->{"teamA"}))
			$match->teamA = YoxTeam::unserialize($json->{"teamA"});
		if (isset($json->{"teamB"}))
			$match->teamB = YoxTeam::unserialize($json->{"teamB"});
		if (isset($json->{"mapPack"}))
			$match->mapPack = $json->{"mapPack"};
		if (isset($json->{"server"}))
			$match->server = YoxServer::unserialize($json->{"server"});



		return $match;
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
class YoxRound
{
	public static function unserialize($json)
	{
	}
	public static function unserializeArray($json)
	{
		$arr = [];
		if (is_array($json))
		{
			foreach ($json as $elem)
			{
				$arr[] = self::unserialize($elem);
			}
		}
		return $arr;
	}
}
abstract class HTTPMethods
{
	const GET = "GET";
	const HEAD = "HEAD";
	const POST = "POST";
	const PUT = "PUT";
	const DELETE = "DELETE";
	const CONNECT = "CONNECT";
	const OPTIONS = "OPTIONS";
	const TRACE = "TRACE";
	const PATCH = "PATCH";
}

/**
 * Client HTTP (Singleton, utiliser ::getInstance())
 */
class HTTPClient
{
	private $host;
	private $port;
	private $header = [];
	private $useragent = "YOXCUP_PLUGIN";
	private $sslverifier = false;
	private $error = "";

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//Constructeur/GetInstance///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	private static $instance = null;

	private function __construct($host, $port, $header)
	{
		$this->host = $host;
		$this->port = $port;
		$this->header = $header;
	}

	/**
	 * @return HTTPClient
	 */
	public static function getInstance($host = "http://127.0.0.1/", $port = 80, $header = [], $forceNew = false)
	{
		if (self::$instance !== null && !$forceNew)
			return self::$instance;

		$i = new HTTPClient($host, $port, $header);
		if (!$forceNew)
			self::$instance = $i;

		return $i;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//Methodes///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	public function sendRequest($page, $method = HTTPMethods::GET, $params = [])
	{
		$ch = curl_init();

		if ($this->useragent !== null)
			curl_setopt($ch, CURLOPT_USERAGENT, $this->useragent);

		if (strpos($this->host, "https") !== false || $this->port = 443)
		{
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, $this->sslverifier);
			curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, $this->sslverifier);
		}

		curl_setopt($ch, \CURLOPT_HTTPHEADER, $this->header);

		switch ($method)
		{
			case HTTPMethods::GET:
				if (count($params) > 0)
				{
					foreach ($params as $key => $value)
					{
						if (strpos($page, "?") === false)
							$page .= "?";
						else
							$page .= "&";

						$page .= "$key=$value";
					}
				}
				break;
			case HTTPMethods::HEAD:
			case HTTPMethods::PUT:
			case HTTPMethods::DELETE:
			case HTTPMethods::CONNECT:
			case HTTPMethods::OPTIONS:
			case HTTPMethods::TRACE:
			case HTTPMethods::PATCH:
				curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
				break;

			case HTTPMethods::POST:
				curl_setopt($ch, CURLOPT_POST, true);
				curl_setopt($ch, CURLOPT_POSTFIELDS, $params);
				break;

			default:
				throw new HTTPClientException("La mÃ©thode $method n'est pas valide");
				break;
		}
		curl_setopt($ch, CURLOPT_URL, $this->host . $page);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		$data = curl_exec($ch);

		$this->error = \curl_error($ch);
		curl_close($ch);
		return $data;
	}

	public function sendJson($page, $json)
	{
		$ch = curl_init();

		if ($this->useragent !== null)
			curl_setopt($ch, CURLOPT_USERAGENT, $this->useragent);

		if (strpos($this->host, "https") !== false || $this->port = 443)
		{
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, $this->sslverifier);
			curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, $this->sslverifier);
		}



		curl_setopt($ch, \CURLOPT_HTTPHEADER, array_merge($this->header, ['Content-Type:application/json']));
		curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
		curl_setopt($ch, CURLOPT_POST, true);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $json);
		curl_setopt($ch, CURLOPT_URL, $this->host . $page);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		$data = curl_exec($ch);

		$this->error = \curl_error($ch);
		curl_close($ch);
		return $data;
	}

	public function addHeader($name, $arg)
	{
		$this->header[] = "$name: $arg";
		//	echo "Add header : " . var_export($this->header, true) . "\n";
	}

	public function clearHeader()
	{
		$this->header = [];
	}

	public function setUserAgent($useragent)
	{
		$this->useragent = $useragent;
	}

	public function getError()
	{
		return $this->error;
	}
}

class HTTPClientException extends \Exception
{
}


/**
 * YoxCup Manager
 *
 * @author  Ankou
 * @version 0.5
 */
class YoxCupPlugin implements CallbackListener, CommandListener, TimerListener, CommunicationListener, Plugin
{
	const _DEBUG_ = true;
	const ID = 170;
	const VERSION = 0.98;
	const NAME = 'YoxCup';
	const AUTHOR = 'Ankou';



	//ManiaLink
	const ML_SERVERS_ID = "YoxCup.ServerList";
	const ML_GOTOBUTTON_ID = "YoxCup.GotoButton";

	//Actions
	const ACT_SERVER_PLAY = "YoxCup.Server.Play";

	//Settings =================================================================

	//Type/Connection
	const SETTING_MATCH_ISLOBBY = 'Is lobby server';
	const SETTING_APIURL = 'API Url';
	const SETTING_SERVER_LOGIN = 'Server login';
	const SETTING_SERVER_TOKEN = 'Server token';

	//Cup/match
	const SETTING_DEFAULT_CUP = 'Default cup id';


	//Interface 
	const SETTING_GOTO_LOBBY = 'Lobby Default Goto link';
	const SETTING_MLGL_POSX = 'Lobby Goto Pos X';
	const SETTING_MLGL_POSY = 'Lobby Goto Pos Y';
	const SETTING_MLGL_WIDTH = 'Lobby Goto Pos Width';
	const SETTING_MLGL_HEIGHT = 'Lobby Goto Pos Height';

	const SETTING_MLSL_POSX = 'ServerList Pos X';
	const SETTING_MLSL_POSY = 'ServerList Pos Y';
	const SETTING_MLSL_HEIGHT = 'ServerList Height';
	const SETTING_MLSL_WIDTH = 'ServerList Width';

	//Debug
	const SETTING_DBG_CALLBACK = '(Debug) log callbacks';
	const SETTING_DBG_EVENTS = '(Debug) log cathed events';
	const SETTING_DBG_LOGVARS = '(Debug) log local variables';




	//Global
	private $isLobby = false;
	private $dbConnected = false;
	private $serverId = -1;
	private $curCupID = -1;
	private $cup;
	private $chatprefix = "";

	//Lobby mode
	/**
	 * @var YoxServer[]
	 */
	private $servers = [];





	//Match mode
	/**
	 * @var YoxServer
	 */
	private $lobbyServer;
	private $curMapID = -1;
	private $curMatchID = -1;
	private $match;
	private $isTeamMatch = true;

	private $nmap = 0;
	private $curTeams;
	private $curTeamA;
	private $curTeamB;

	private $roundRunning = false;

	private $isWarmUp = false;


	private $curRound;
	private $curRoundsLimit;
	private $curMapLimit;
	private $matchStarted = false;
	private $matchCurMapNb = 0;
	private $matchPaused = false;
	private $matchMapPack = null;
	private $startOnNextMap = false;


	private $roundWinByA = 0;
	private $roundWinByB = 0;
	private $mapWinByA = 0;
	private $mapWinByB = 0;

	//Stats globals ==========================================================

	//Array [login => [weapon => nbShoot]]
	private $shoots;
	//Array [login => nbHit]
	private $hits;
	//Array [login => nbHit]
	private $hitted;
	//Array [login => nbKills]
	private $kills;
	//Array [login => nbDeaths]
	private $death;


	//Stats elite ============================================================
	private $tmpVictoryType = 0;
	private $tmpAttacker = "";

	//Array [login => [winType]]
	private $attacksWin;


	//debug
	private $dbgCallbacks = false;
	private $dbgEvents = false;
	private $dbgVars = false;

	/**
	 * @var HTTPClient client
	 */
	private $apiClient = null;

	private $apiToken = null;


	private $forceAdminSpec = true;
	/**
	 * Private Properties
	 */
	/** @var ManiaControl $maniaControl */
	private $maniaControl = null;



	/**
	 * @see \ManiaControl\Plugins\Plugin::load()
	 */
	public function load(ManiaControl $maniaControl)
	{
		$this->maniaControl = $maniaControl;


		//Commands
		$this->maniaControl->getCommandManager()->registerCommandListener('setcupid', $this, 'onCommandSetCupId', true, 'Sets the cupID.');
		$this->maniaControl->getCommandManager()->registerCommandListener('setmatchid', $this, 'onCommandSetMatchId', true, 'Sets the matchID.');
		$this->maniaControl->getCommandManager()->registerCommandListener('startmatch', $this, 'onCommandStartMatch', true, 'Start the match.');
		$this->maniaControl->getCommandManager()->registerCommandListener('listmatch', $this, 'onCommandListMatch', true, 'List cup matchs');
		$this->maniaControl->getCommandManager()->registerCommandListener('pushmaps', $this, 'onCommandShareMaps', true, 'Share server map list with YoxCup database.');
		$this->maniaControl->getCommandManager()->registerCommandListener('getmaps', $this, 'onGetMaps', true, 'Download maps from YoxCup database (over MX)');
		$this->maniaControl->getCommandManager()->registerCommandListener('showvars', $this, 'onCommandShowVars', true, 'Show yoxcup variables');
		$this->maniaControl->getCommandManager()->registerCommandListener('clearmatch', $this, 'onCommandClearMatch', true, 'Clear locals variables about the current match');
		$this->maniaControl->getCommandManager()->registerCommandListener('pause', $this, 'onCommandPause', true, 'Set pause');
		$this->maniaControl->getCommandManager()->registerCommandListener('yoxhelp', $this, 'onCommandYoxHelp', true, 'Show yoxcup commands');

		//test update map from MX
		//$this->maniaControl->getMapManager()->getMXManager()->fetchMaplistByMixedUidIdString("105644");

		//===========================================================================================================================
		//Callbacks==================================================================================================================
		//===========================================================================================================================



		//Debug
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_MODESCRIPTCALLBACK, $this, 'debugCallBack');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_MODESCRIPTCALLBACKARRAY, $this, 'debugCallBack');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::SCRIPTCALLBACK, $this, 'debugCallBack');
		$this->maniaControl->getCallbackManager()->registerCallbackListener("CallBacks.ScriptCallbackArray", $this, 'debugCallBack');

		//Common 
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::BEGINMAP, $this, 'cb_mp_beginMap');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::ENDMAP, $this, 'cb_mp_endMap');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(SettingManager::CB_SETTING_CHANGED, $this, 'updateSettings');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_PLAYERMANIALINKPAGEANSWER, $this, 'handleManialinkPageAnswer');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(PlayerManager::CB_PLAYERCONNECT, $this, 'cb_mp_playerConnect');




		//ShootMania Common
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::SM_ONSHOOT, $this, "cb_sm_onshoot");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::SM_ONHIT, $this, "cb_sm_onhit");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("ShootMania.Event.OnArmorEmpty", $this, "cb_sm_onarmorempty");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("ShootMania.Event.OnCapture", $this, "cb_sm_oncapture");


		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::SM_SCORES, $this, "cb_sm_scores");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::MP_WARMUP_END, $this, "cb_mp_warmupEnd");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::MP_WARMUP_START, $this, "cb_mp_warmupStart");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener(CallBacks::MP_PAUSE_STATUS, $this, "cb_mp_pause");


		//CTF  
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("CTF.EndMap.Scores", $this, "cb_ctf_endscores");

		//Elite
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("Shootmania.Elite.StartTurn", $this, "cb_elite_startTurn");
		$this->maniaControl->getCallbackManager()->registerScriptCallbackListener("Shootmania.Elite.EndTurn", $this, "cb_elite_endTurn");







		//===========================================================================================================================
		//settings===================================================================================================================
		//===========================================================================================================================

		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MATCH_ISLOBBY, false);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_APIURL, "https://api.yoxclan.fr/");
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_SERVER_TOKEN, "serverToken", "YoxCup API authentication token");
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_DEFAULT_CUP, 0);


		//ServerList manialink settings
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLSL_POSX, -140);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLSL_POSY, -10);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLSL_WIDTH, 40);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLSL_HEIGHT, 70);


		//goto lobby button manialink settings
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_GOTO_LOBBY, "maniaplanet://#qspectate=yourserver@TitlePack", "Maniaplanet link to join the lobby");
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLGL_POSX, 156);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLGL_POSY, -9);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLGL_WIDTH, 6);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_MLGL_HEIGHT, 6);

		//Debug
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_DBG_CALLBACK, false);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_DBG_EVENTS, false);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_DBG_LOGVARS, false);


		$this->isLobby = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MATCH_ISLOBBY);
		$this->dbgCallbacks = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_CALLBACK);
		$this->dbgEvents = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_EVENTS);
		$this->dbgVars = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_LOGVARS);
		$this->authenticateServer();
		$this->buildManialink();
	}


	public function hideManialinks()
	{
		$this->maniaControl->getManialinkManager()->hideManialink(self::ML_SERVERS_ID);
		$this->maniaControl->getManialinkManager()->hideManialink(self::ML_GOTOBUTTON_ID);
	}

	public function updateSettings()
	{
		$this->isLobby = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MATCH_ISLOBBY);
		$this->dbgCallbacks = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_CALLBACK);
		$this->dbgEvents = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_EVENTS);
		$this->dbgVars = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DBG_LOGVARS);
		$this->isLobby = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MATCH_ISLOBBY);
		$this->hideManialinks();
		$this->authenticateServer();
		$this->buildManialink();
	}


	private function authenticateServer()
	{
		$login = $this->maniaControl->getClient()->getMainServerPlayerInfo()->login;
		$token = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_SERVER_TOKEN);
		$defCup = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_DEFAULT_CUP);
		$apiUrl = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_APIURL);

		$this->apiClient = HTTPClient::getInstance($apiUrl, 443);
		$r = $this->apiClient->sendRequest("cup/serverconnect", HTTPMethods::POST, [
			"login" => $login,
			"token" => $token
		]);


		$data = json_decode($r);
		if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("[YoxCupPlugin]>Auth error : " . $data->{"Message"});
			return;
		}



		if (isset($data->{"Token"}))
		{
			$this->apiToken = $data->{"Token"};
			$this->serverId = intval($data->{"id"});
			$this->apiClient->clearHeader();
			$this->apiClient->addHeader("x-api-key", $this->apiToken);

			$this->maniaControl->getChat()->sendSuccessToAdmins("[YoxCup]>Authentication : OK");
			$this->dbConnected = true;
			if ($defCup != 0)
			{
				$this->setCupId($defCup);
			}
			//	$this->serverId = intval($r["id"]);
			return;
		}
	}


	public function logEvent($message)
	{
		if ($this->dbgEvents)
			$this->maniaControl->getChat()->sendInformation("Event -> " . $message);
	}


	//=================================================================================================================================================================
	//==[OnCallback]===================================================================================================================================================
	//=================================================================================================================================================================




	//Common ====================================
	public function cb_mp_endMap($struct)
	{
		//echo var_export($struct, true);
	}


	public function cb_sm_onshoot($struct)
	{
		if ($this->match == null || !($this->matchStarted) || !($this->roundRunning)  || $this->isWarmUp)
		{
			return;
		}
		$struct = json_decode($struct[1][0]);
		if (isset($this->shoots[$struct->{"shooter"}][$struct->{"weapon"}]))
			$this->shoots[$struct->{"shooter"}][$struct->{"weapon"}] += 1;
		else
			$this->shoots[$struct->{"shooter"}][$struct->{"weapon"}] = 1;
	}
	public function cb_sm_onhit($struct)
	{
		if ($this->match == null || !($this->matchStarted) || !($this->roundRunning) || $this->isWarmUp)
		{
			return;
		}
		$struct = json_decode($struct[1][0]);

		if (isset($this->hits[$struct->{"shooter"}][$struct->{"weapon"}]))
			$this->hits[$struct->{"shooter"}][$struct->{"weapon"}] += 1;
		else
			$this->hits[$struct->{"shooter"}][$struct->{"weapon"}] = 1;

		if (isset($this->hitted[$struct->{"victim"}][$struct->{"weapon"}]))
			$this->hitted[$struct->{"victim"}][$struct->{"weapon"}] += 1;
		else
			$this->hitted[$struct->{"victim"}][$struct->{"weapon"}] = 1;
	}
	public function cb_sm_onarmorempty($struct)
	{
		//echo var_export($struct, true);
	}
	public function cb_sm_oncapture($struct)
	{
		//echo var_export($struct, true);
	}

	public function cb_mp_beginMap(Map $map)
	{
		if ($this->cup == null)
			return;


		switch ($this->cup->gameMode)
		{
			case "CTF Rockets": //CTF====================================================================================

				if (!$this->matchStarted)
				{
					if ($this->startOnNextMap)
					{
						$this->matchStarted = true;
						$this->startOnNextMap = false;
						$this->maniaControl->getChat()->sendInformation("Match started !");
					}
					return;
				}



				if ($this->curMapID > $this->matchCurMapNb)
				{
					$this->matchStarted = false;
					$this->maniaControl->getChat()->sendInformation("Match finished !");
					return;
				}
				try
				{
					if (!$this->matchMapPack[$this->curMapID]["uid"] === $map->uid)
					{
						$this->matchPaused = true;
						//			$this->maniaControl->getChat()->sendError("this map is not in map pack ! match paused");
						$this->maniaControl->getClient()->setWarmUp(true);
						$this->maniaControl->getClient()->forceEndRound();
					}
				}
				catch (\Exception $e)
				{
					$this->matchPaused = true;
					//		$this->maniaControl->getChat()->sendError("Error in MapPack ! match paused");
					$this->maniaControl->getClient()->setWarmUp(true);
					$this->maniaControl->getClient()->forceEndRound();
				}
				break;

			case "Elite": //Elite =======================================================================================================
				//Reset infos
				/*$this->curMatchID = -1;
				$this->matchPaused = false;
				$this->matchStarted = false;**/
				$this->roundWinByA = 0;
				$this->roundWinByB = 0;
				$this->hits = [];
				$this->hitted = [];
				$this->kills = [];

				break;
		}
	}

	public function cb_mp_warmupEnd()
	{
		if ($this->matchPaused)
		{
			$this->matchPaused = false;
		}
		else
		{
			//Start map (elite)
			if ($this->cup != null && $this->cup->gameMode == "Elite" && $this->match != null && $this->matchStarted)
			{
			}
		}
		$this->isWarmUp = false;
	}

	public function cb_mp_warmupStart()
	{
		$this->isWarmUp = true;
	}

	public function cb_mp_pause($struct = null)
	{
		$this->matchPaused = true;
		$this->shoots = [];
		$this->hits = [];
		$this->hitted = [];
	}

	public function cb_sm_scores($struct)
	{
		if ($this->cup == null || $this->match == null || !($this->matchStarted))
			return;


		$struct = json_decode($struct[1][0]);
		switch ($this->cup->gameMode)
		{
			case "Elite": //Elite =======================================================================================================
				if (isset($struct->{"section"}) && !($this->isWarmUp))
				{
					$this->logEvent("Scores : " . $struct->{"section"});

					switch ($struct->{"section"})
					{
						case "EndTurn":
							$this->logEvent("Push round result to API ...");
							$winner = $struct->{"winnerteam"};
							//Save result
							if ($winner != 0)
							{
								$idWinner = $this->curTeamA->id;
								$idLoser = $this->curTeamB->id;
								$this->roundWinByA++;
							}
							else
							{
								$idWinner = $this->curTeamB->id;
								$idLoser = $this->curTeamA->id;
								$this->roundWinByB++;
							}

							$data = json_encode((object) [
								"Type" => "Data",
								"Values" => (object) [
									"idcup" => $this->cup->id,
									"idmatch" => $this->match->nMatch,
									"uidmap" => $this->maniaControl->getMapManager()->getCurrentMap()->uid,
									"nmap" => $this->nmap,
									"idround" => $this->curRound,
									"type" => 1,
									//1 = round win
									"idwinner" => $idWinner,
									"idloser" => $idLoser,
								]
							]);

							$r = $this->apiClient->sendJson("results/push", $data);
							$this->curRound++;




							//Save stats========================================================================
							//Build context
							$this->logEvent("Push round stats to API ...");
							$context = (object) [
								"idcup" => $this->cup->id,
								"idmatch" => $this->match->nMatch,
								"uidmap" => $this->maniaControl->getMapManager()->getCurrentMap()->uid,
								"nmap" => $this->nmap,
								"idround" => $this->curRound,
							];

							//Build data part
							$stats = [];
							$stats = array_merge($stats, $this->util_prepareStat($this->shoots, "Shoot"));
							$stats = array_merge($stats, $this->util_prepareStat($this->hits, "Hits"));
							$stats = array_merge($stats, $this->util_prepareStat($this->hitted, "Hitted"));
							$stats = array_merge($stats, $this->util_prepareStat([
								$this->tmpAttacker => $this->tmpVictoryType
							], "WinType"));


							//Build final Json

							$json = json_encode((object) [
								"Type" => "Data",
								"Values" => (object) [
									"Infos" => $context,
									"Data" => $stats
								]
							]);

							$r = $this->apiClient->sendJson("stats/push", $json);



							/*	echo "wintype : " . var_export($this->tmpVictoryType, true) . "\n";
							echo "attacker : " . var_export($this->tmpAttacker, true) . "\n";*/

							//reset data
							$this->shoots = [];
							$this->hits = [];
							$this->hitted = [];
							break;

						case "EndMap":
							$this->logEvent("Push map result to API ...");
							$winner = $struct->{"winnerteam"};
							//Save result
							if ($winner != 0)
							{
								$idWinner = $this->curTeamA->id;
								$idLoser = $this->curTeamB->id;
								$this->mapWinByA++;
							}
							else
							{
								$idWinner = $this->curTeamB->id;
								$idLoser = $this->curTeamA->id;
								$this->mapWinByB++;
							}

							$data = json_encode((object) [
								"Type" => "Data",
								"Values" => (object) [
									"idcup" => $this->cup->id,
									"idmatch" => $this->match->nMatch,
									"uidmap" => $this->maniaControl->getMapManager()->getCurrentMap()->uid,
									"nmap" => $this->nmap,
									"idround" => 999,
									"type" => 4,
									//4 = map win
									"idwinner" => $idWinner,
									"idloser" => $idLoser,
								]
							]);

							$r = $this->apiClient->sendJson("results/push", $data);
							$this->curRound = 1;
							//Check if match finished
							if ($this->nmap == $this->cup->nMapPerMatch || $this->mapWinByA == (intval($this->cup->nMapPerMatch / 2) + 1) || $this->mapWinByB == (intval($this->cup->nMapPerMatch / 2) + 1))
							{
								$message = "Match finished ! Winner : ";
								if ($this->mapWinByA > $this->mapWinByB)
								{
									$message .= $this->curTeamA->name;
								}
								else
								{
									$message .= $this->curTeamB->name;
								}

								$this->maniaControl->getChat()->sendInformation($message);
								$this->matchStarted = false;
								$this->match = null;
							}
							else
							{
								$this->logEvent("Nmap : " . $this->nmap . " -> " . $this->nmap + 1);
								$this->nmap++;
								$this->curMapID++;
							}
							//RAZ
							$this->roundWinByA = 0;
							$this->roundWinByB = 0;




							break;
					}
				}



				break;
		}
	}


	public function util_getPlayerObject($login, &$idTeam)
	{
		$idTeam = 0;
		if ($this->curTeamA != null && $this->curTeamB != null)
		{
			foreach ($this->curTeamA->players as $player)
			{
				if ($player->login == $login)
				{
					$idTeam = $this->curTeamA->id;
					return $player;
				}
			}
			foreach ($this->curTeamB->players as $player)
			{
				if ($player->login == $login)
				{
					$idTeam = $this->curTeamB->id;
					return $player;
				}
			}
		}
		return null;
	}


	public function util_prepareStat($data, $type)
	{
		$stats = [];
		foreach ($data as $login => $sub)
		{
			$idPlayer = 0;
			$player = $this->util_getPlayerObject($login, $idTeam);
			if ($player != null)
			{
				$idPlayer = $player->id;
			}
			if (is_array($sub))
			{
				foreach ($sub as $key => $value)
				{
					$stats[] = (object) [
						"type" => $type,
						"idPlayer" => $idPlayer,
						"idTeam" => $idTeam,
						"subType" => $key,
						"value" => $value
					];
				}
			}
			else
			{
				$stats[] = (object) [
					"type" => $type,
					"idPlayer" => $idPlayer,
					"idTeam" => $idTeam,
					"subType" => 0,
					"value" => $sub
				];
			}
		}
		return $stats;
	}


	/**
	 * On playerConnect callback
	 * TODO: rename method
	 */
	public function cb_mp_playerConnect(Player $player)
	{
		echo "player_connect\n";
		if ($this->cup != null && $this->match != null)
		{
			$officialPlayer = false;
			//$this->curTeamA->players;
			foreach($this->curTeamA->players as $p)
			{
				if($p->login == $player->login)
				{
					$this->maniaControl->getClient()->forcePlayerTeam($player->login, 0);
					$officialPlayer = true;
					echo "find A\n";
					break;
				}
			}
			foreach($this->curTeamB->players as $p)
			{
				if($p->login == $player->login)
				{
					$this->maniaControl->getClient()->forcePlayerTeam($player->login, 1);
					$officialPlayer = true;
					echo "find B\n";
					break;
				}
			}
			if(!($officialPlayer))
			{
				echo "unofficial, level : " . $player->authLevel . " isspec : " . var_export($player->isSpectator,true) . "\n";
				if(($player->authLevel < 3 || $this->forceAdminSpec) && !$player->isSpectator)
					$this->maniaControl->getClient()->forceSpectator($player->login, 1);
			}

			
		
		
		


		}


		return;
		$this->buildManialink($player->login);


		//check teams
		if ($this->curMatchID != -1 && $this->match != null)
		{
			$this->logEvent("CheckPlayerTeam");
			$isPl = false;
			$a = $this->getTeamPlayers($this->match->teamA);
			$b = $this->getTeamPlayers($this->match->teamB);

			if ($this->util_teamHasPlayer($a, $player->login))
			{
				$this->maniaControl->getChat()->sendInformation($player->nickname . '$z forced in A');
				$this->maniaControl->getClient()->forcePlayerTeam($player, 0);
			}
			else
			{
				if ($this->util_teamHasPlayer($b, $player->login))
				{
					$this->maniaControl->getChat()->sendInformation($player->nickname . '$z forced in B');
					$this->maniaControl->getClient()->forcePlayerTeam($player, 1);
				}
				else
				{
					$this->maniaControl->getChat()->sendInformation($player->nickname . '$z forced in Spec');
					$this->maniaControl->getClient()->forceSpectator($player, 1);
				}
			}
		}
	}

	/**
	 * @deprecated Only for debug
	 */
	public function debugCallBack($struct)
	{
		if (is_array($struct))
		{
			if ($struct[0] === "ManiaPlanet.ModeScriptCallbackArray" && $struct[1][0] != "Shootmania.UI.Properties")
			{
				if ($struct instanceof OnShootStructure)
				{
					echo "OnShootStructure\n";
				}
				else
				{
					//echo var_export($struct, true) . "\n";
					if (isset($struct[1][0]) && is_string($struct[1][0]) && $this->dbgCallbacks)
					{
						$this->maniaControl->getChat()->sendInformation("CB -> " . $struct[1][0]);
					}
				}
			}

			$cbName = $struct[1][0];
			switch ($cbName)
			{
				case "Shootmania.Scores":


					break;
				case "CTF.EndMap.Scores":
					if ($this->matchStarted)
					{
						$json = json_decode($struct[1][1]);

						foreach ($json->{"Players"} as $player)
						{
							$this->insertPlayerStat($player);
						}
					}
					break;
			}
		}
	}

	//Elite ==============================================================================
	public function cb_elite_startTurn($struct)
	{
		if (!($this->isWarmUp))
		{
			$this->roundRunning = true;
			if (isset($struct[1][0]))
			{
				$struct = json_decode($struct[1][0]);
				$this->tmpAttacker = $struct->{"attacker"};
				$this->logEvent("Start turn, attack : " . $this->tmpAttacker);
			}
		}
	}

	public function cb_elite_endTurn($struct = "")
	{
		if (isset($struct[1][0]))
		{
			$struct = json_decode($struct[1][0]);
			if (isset($struct->{"victorytype"}))
				$this->tmpVictoryType = $struct->{"victorytype"};
		}
		$this->roundRunning = false;
	}







	//CTF=================================================================================
	public function cb_ctf_endscores($struct)
	{
		if ($this->matchStarted)
		{
			$json = json_decode($struct[1][1]);

			foreach ($json->{"Players"} as $player)
			{
				$this->insertPlayerStat($player);
			}
		}
	}







	//=================================================================================================================================================================
	//==[OnCommands]===============================================================================================================================================================
	//=================================================================================================================================================================

	public function onCommandPause(array $chatCallback, Player $player)
	{
		//Assertions --------------------------------------------------

		if (!$this->checkPrivileges($player))
			return;

		if (!$this->matchStarted)
		{
			//	$this->maniaControl->getChat()->sendError("Cannot pause an unstarted match", $player->login);
			return;
		}
		if ($this->matchPaused)
		{
			//	$this->maniaControl->getChat()->sendError("Match already paused", $player->login);
			return;
		}


		//Pause --------------------------------------------------------

		$this->maniaControl->getClient()->restartMap(true); //restart la map sans effacer les scores. (le mode compÃ©titif doit Ãªtre actif)
		/*		$this->maniaControl->getClient()->setWarmUp(true);
		$this->maniaControl->getClient()->forceEndRound();*/
		//	$this->maniaControl->getChat()->sendInformation($player->nickname . '$z$fff : Set pause');
	}

	public function onCommandClearMatch(array $chatCallback, Player $player)
	{
		$this->match = null;
		$this->curMapID = -1;
		$this->curMatchID = -1;
		$this->mapWinByA = 0;
		$this->mapWinByB = 0;
		$this->roundWinByA = 0;
		$this->roundWinByB = 0;
		$this->nmap = 0;
		$this->maniaControl->getClient()->setTeamInfo("Blue", 0.66, "", "Red", 0.99, "");
		$this->maniaControl->getChat()->sendInformation("Match cleared !", $player->login);
		//$this->maniaControl->getChat()->sendInformationToAdmins("Match cleared !");
	}
	public function onCommandShowVars(array $chatCallback, Player $player)
	{
		if (!$this->checkPrivileges($player))
			return;


		$data = "";
		$data .= "Is Lobby : " . var_export($this->isLobby, true) . "\n";
		$data .= "Server ID : " . var_export($this->serverId, true) . "\n";
		$data .= "Cur Map ID : " . var_export($this->curMapID, true) . "\n";
		$data .= "Cur Match ID : " . var_export($this->curMatchID, true) . "\n";
		//$data .= "Cur Map Pack : " . 	 	var_export($this->curMapPack, true) . "\n";
		$data .= "Cur Round : " . var_export($this->curRound, true) . "\n";
		$data .= "Cur Round Limit : " . var_export($this->curRoundsLimit, true) . "\n";
		$data .= "Cur Map Limit : " . var_export($this->curMapLimit, true) . "\n";
		$data .= "Match Started : " . var_export($this->matchStarted, true) . "\n";
		$data .= "Match Paused : " . var_export($this->matchPaused, true) . "\n";
		$data .= "Warmup State : " . var_export($this->isWarmUp, true) . "\n";
		//	$data .= "Cur team A : " . 		 	var_export($this->curTeamA, true) . "\n";
		//$data .= "Cur team A players : " . 	var_export(count($this->curTeamAplayers, true)) . "\n";
		//	$data .= "Cur team B: " . 		 	var_export($this->curTeamB, true) . "\n";
		//$data .= "Cur team B players : " . 	var_export(count($this->curTeamBplayers, true)) . "\n";
		$data .= "Match Cur Map Nb : " . var_export($this->matchCurMapNb, true) . "\n";
		$data .= "Start On Next Map : " . var_export($this->startOnNextMap, true) . "\n";








		$this->maniaControl->getChat()->sendInformation($data, $player->login);
	}

	public function onCommandYoxHelp(array $chatCallback, Player $player)
	{
		if (!$this->checkPrivileges($player))
			return;

		$message = "$11fYoxCupCommands : \n";
		/*	$message .= '$999- User Commands$aff' . "\n";
		$message .= '//pause       (ask pause)' . "\n";
		$message .= '//request     (ask for admin)' . "\n";*/
		$message .= '$999- Match management$aa0' . "\n";
		$message .= '//setcupid          #id' . "\n";
		$message .= '//listmatch' . "\n";
		$message .= '//setmatchid      #id' . "\n";
		$message .= '//addmatchmap #id' . "\n";
		$message .= '//pausematch' . "\n";
		$message .= '//startmatch' . "\n";
		$message .= '$999- SuperAdmin/Dangerous Commands$f00' . "\n";
		$message .= '//clearmatch   (For rematch, reset all data saved about the current match)' . "\n";
		$message .= '//sharemaps   (Share local server maps with YoxDatabase)' . "\n";
		$message .= '//forcestop      (Force the end of the match in case of bug)' . "\n";
		$message .= '$999- Debug Commands$f00' . "\n";
		$message .= '//showvars' . "\n";





		$this->maniaControl->getChat()->sendInformation($message, $player->login);
	}


	/**
	 * Command //startmatch for admins
	 * 
	 * @param array			$chatCallback
	 * @param \ManiaControl\Players\Player $player
	 */
	public function onCommandStartMatch(array $chatCallback, Player $player)
	{
		//Assertions ===================================================================================
		if (!$this->checkPrivileges($player))
			return;

		if ($this->matchStarted)
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("Cannot start a match already started");
			return;
		}

		if ($this->curMatchID === -1 || $this->curCupID === -1 || $this->isLobby)
		{
			//$this->maniaControl->getChat()->sendErrorToAdmins("Cannot start match, check lobby status, cupid and matchid");
			return;
		}

		if ($this->curCupID == 0 || $this->curMatchID == 0 || $this->curMapLimit == 0)
		{
			//$this->maniaControl->getChat()->sendErrorToAdmins("Cannot start match ! (Check config)");
			return;
		}


		//Start ==============================================================================================

		/*	$this->matchCurMapNb = $this->getMapPack($player);
		if ($this->maniaControl->getMapManager()->getCurrentMap()->uid === $this->matchMapPack[0]["uid"])
		{
		$this->maniaControl->getClient()->restartMap();
		}
		else
		{
		$this->maniaControl->getClient()->nextMap();
		}*/
		//		$this->maniaControl->getChat()->sendInformation("Match : " . $this->curTeamA->name . '$z vs ' . $this->curTeamB->name . '$z will start after restart !');
		//$this->startOnNextMap = true;

		$this->maniaControl->getChat()->sendInformation("Match : " . $this->curTeamA->name . '$z vs ' . $this->curTeamB->name . '$z start at warmup end !');
		$this->matchStarted = true;
		$this->curMapID = 1;
		$this->curRound = 1;
		$this->roundWinByA = 0;
		$this->roundWinByB = 0;
		$this->nmap = 1;
	}

	public function onCommandListMatch(array $chatCallback, Player $player)
	{
		if (!$this->checkPrivileges($player) || $this->cup == null)
			return;

		try
		{

			$r = $this->apiClient->sendRequest("cup/" . $this->curCupID . "/matches", HTTPMethods::GET, []);
			$data = json_decode($r);

			if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
			{
				return -1;
				echo "YoxCup Api Error : " . $data->{"Message"} . "\n";
			}

			$matches = YoxMatch::unserializeArray($data->{"Values"});
			foreach ($matches as $match)
			{
				$this->maniaControl->getChat()->sendInformation("Match : " . $match->nMatch . " -> " . $match->teamA->name . '$z VS ' . $match->teamB->name, $player->login);
			}
		}
		catch (Exception $e)
		{
			return -1;
			echo $e->getMessage() . "\n";
		}
	}


	/**
	 * Command //SetCupId for admins
	 * 
	 * @param array			$chatCallback
	 * @param \ManiaControl\Players\Player $player
	 */
	public function onCommandSetCupId(array $chatCallback, Player $player)
	{
		if (!$this->checkPrivileges($player))
			return;

		if ($this->matchStarted)
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("cannot define cupid, a match is already in progress");
			return;
		}

		$text = $chatCallback[1][2];
		$text = explode(" ", $text);
		$this->setCupId($text[1], $player);

		//$lastmatches = $this->getMatchesList(3);


	}

	public function setCupId($id, $player = null)
	{
		if (!is_numeric($id))
		{
			$message = $this->chatprefix . '$<run this command with an index$>' . "\n";
			$this->maniaControl->getChat()->sendSuccess($message, $player);
			//$this->maniaControl->getChat()->sendError($this->chatprefix . 'For the moment, only point recovery is supported, you have to manage maps and rounds manually');
			return;
		}
		if ($this->isLobby)
		{
			$cup = $this->getCup($id);

			if ($cup !== null && $cup !== false)
			{
				$this->logEvent("Store cupId");
				$this->curCupID = intval($id);
				$this->hideManialinks();
				$this->buildManialink();
			}

			//$this->maniaControl->getChat()->sendErrorToAdmins("cannot set cupid on lobby server");
			return;
		}




		$cup = $this->getCup($id);
		if ($cup == null)
		{
			$this->logEvent("setCupId() -> cup == null");
		}
		else
		{
			$this->logEvent("setCupId() -> cup == OK");
		}
		if ($cup !== null && $cup !== false)
		{
			$script = $this->maniaControl->getClient()->getGameInfos()["CurrentGameInfos"]->scriptName;
			if (strpos($script, "\\") != false)
				$script = substr($script, 0, strpos($script, "\\"));
			else
				$script = substr($script, 0, strpos($script, "/"));

			if ($script != $cup->gameMode)
			{
				//	$this->maniaControl->getChat()->sendError($this->chatprefix . 'Bad game mode (local : ' . $script .  "; Cup : " . $cup->gameMode . ")");
				//	return;
			}



			$this->curCupID = intval($id);
			$this->cup = $cup;
			$this->curMapLimit = $cup->nMapPerMatch;




			$this->maniaControl->getChat()->sendSuccess("Cup : " . $cup->name, null);
		}
		else
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("La cup " . $id . " n'existe pas !");
		}
	}




	/**
	 * Command //SetCupId for admins
	 * 
	 * @param array			$chatCallback
	 * @param \ManiaControl\Players\Player $player
	 */
	public function onCommandSetMatchId(array $chatCallback, Player $player)
	{
		if (!($this->checkPrivileges($player)))
			return;

		if ($this->matchStarted)
		{
			$this->maniaControl->getChat()->sendErrorToAdmins("Cannot define matchid, a match is already in progress");
			return;
		}

		if ($this->curCupID === -1 || $this->isLobby || $this->cup == null)
		{
			$this->maniaControl->getChat()->sendErrorToAdmins("Cannot define match, check lobby status and cupid");
			return;
		}


		$text = $chatCallback[1][2];
		$text = explode(" ", $text);

		if (is_numeric($text[1]))
		{
			try
			{
				$this->match = null;
				$this->curMapID = -1;
				$this->curMatchID = -1;
				$this->mapWinByA = 0;
				$this->mapWinByB = 0;
				$this->roundWinByA = 0;
				$this->roundWinByB = 0;
				$this->nmap = 0;
				$match = $this->getMatch($text[1]);
				//	echo var_export($match, true);
				if ($match !== null && $match !== false)
				{
					$this->match = $match;
					$this->curMatchID = $text[1];
					$this->curTeamA = $match->teamA;
					$this->curTeamB = $match->teamB;


					$this->maniaControl->getChat()->sendSuccess("Match : " . $match->id, null);

					$this->maniaControl->getClient()->setTeamInfo($this->curTeamA->name, 0.66, "", $this->curTeamB->name, 0.99, "");



					$teamA = '$00fTeam ' . $this->curTeamA->name . ' $z$fff: ';
					$playersA = $this->getTeamPlayers($this->curTeamA->id);
					$this->curTeamA->players = $playersA;

					$players = $this->maniaControl->getClient()->getPlayerList();
					foreach ($playersA as $player)
					{
						$teamA .= $player->username . '$z$fff, ';
						foreach($players as $p)
						{
							if($p->login == $player->login && $p->teamId != 0)
								$this->maniaControl->getClient()->forcePlayerTeam($player->login, 0);
						}
					}
					$this->maniaControl->getChat()->sendInformation($teamA);


					$teamB = '$f00Team ' . $this->curTeamB->name . ' $z$fff: ';
					$playersB = $this->getTeamPlayers($this->curTeamB->id);
					$this->curTeamB->players = $playersB;
					foreach ($playersB as $player)
					{
						$teamB .= $player->username . '$z$fff, ';
						foreach($players as $p)
						{
							if($p->login == $player->login && $p->teamId != 1)
								$this->maniaControl->getClient()->forcePlayerTeam($player->login, 1);
						}
					}

					$this->maniaControl->getChat()->sendInformation($teamB);




					//Settings=============================
					switch ($this->cup->gameMode)
					{
						case "CTF Rockets":
							$settings = [
								"S_TurnsToWin" => (int) $this->cup->nRoundPerMatch,
								"S_NbPlayersPerTeamMin" => (int) $this->cup->nPlayerByTeam,
								"S_NbPlayersPerTeamMax" => (int) $this->cup->nPlayerByTeam,
								"S_EnableCompetitive" => true
							];
							$this->maniaControl->getClient()->setModeScriptSettings($settings);
							break;

						case "Elite":
							//TODO: Elite settings
							break;
					}
				}
				else
				{
					//	$this->maniaControl->getChat()->sendErrorToAdmins("Le match " . $text[1] . "n'existe pas !");
				}
			}
			catch (Exception $e)
			{
				//$this->maniaControl->getChat()->sendErrorToAdmins($e->getMessage());
			}
		}
		else
		{
			//$lastmatches = $this->getMatchesList(3);

			$message = $this->chatprefix . '$<run this command with an index$>' . "\n";
			$this->maniaControl->getChat()->sendSuccess($message, $player);
			//	$this->maniaControl->getChat()->sendError($this->chatprefix . 'For the moment, only point recovery is supported, you have to manage maps and rounds manually');
		}
	}



	public function onCommandShareMaps(array $chatCallback, Player $player)
	{
		if (!$this->checkPrivileges($player))
			return;


		$this->maniaControl->getMapManager()->handleOnInit();
		$maplist = $this->maniaControl->getMapManager()->getMaps();
		//echo var_dump($maplist);
		foreach ($maplist as $map)
		{
			$this->insertMapInYoxCup($map);
		}
	}






	//=================================================================================================================================================================
	//==Get/set infos from API=========================================================================================================================================
	//=================================================================================================================================================================

	/**
	 * Insert stats in DB
	 * TODO: Make this more abstract & generic
	 * @param object $player (JsonObject)
	 */
	private function insertPlayerStat($player)
	{
		if (!($this->matchStarted))
		{
			return;
		}

		$id = $this->getPlayerId($player->{"Login"});
		if ($player->{"Clan"} === 1)
			$idTeam = $this->curTeamA["id"];
		else
			$idTeam = $this->curTeamB["id"];


		if ($id === -1) //Joueur inexistant dans la BDD
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("Le joueur " . $player->{"Login"} . " n'est pas dÃ©clarÃ©.");
			return;
		}

		//TODO: write API call
		$sth = $this->db->prepare("INSERT INTO yox_stats (type, idMatch, idCup, idPlayer, idTeam, idRound, value) VALUES (:type, :match, :idcup, :player, :team, :round, :val);");

		$sth->execute([
			"type" => "Hits",
			"match" => $this->curMatchID,
			"idcup" => $this->curCupID,
			"player" => $id,
			"team" => $idTeam,
			"round" => $this->curRound,
			"val" => $player->{"Hits"}
		]);

		$sth->execute([
			"type" => "CapPoints",
			"match" => $this->curMatchID,
			"idcup" => $this->curCupID,
			"player" => $id,
			"team" => $idTeam,
			"round" => $this->curRound,
			"val" => $player->{"CapPoints"}
		]);

		$sth->execute([
			"type" => "AtkPoints",
			"match" => $this->curMatchID,
			"idcup" => $this->curCupID,
			"player" => $id,
			"team" => $idTeam,
			"round" => $this->curRound,
			"val" => $player->{"AtkPoints"}
		]);

		$sth->execute([
			"type" => "DefPoints",
			"match" => $this->curMatchID,
			"idcup" => $this->curCupID,
			"player" => $id,
			"team" => $idTeam,
			"round" => $this->curRound,
			"val" => $player->{"DefPoints"}
		]);

		$sth->execute([
			"type" => "Combo",
			"match" => $this->curMatchID,
			"idcup" => $this->curCupID,
			"player" => $id,
			"team" => $idTeam,
			"round" => $this->curRound,
			"val" => $player->{"Combo"}
		]);
	}

	/**
	 * @param Map $map 
	 */
	private function insertMapInYoxCup($map)
	{
		//TODO: Make API call
		try
		{
			$db = $this->db;

			$sth = $db->prepare("INSERT INTO yox_maps (mxid, uid,name,author,file) VALUE (:mxid, :uid, :name, :author, :file);");
			$sth->execute([
				"mxid" => $map->mx === null ? null : $map->mx->id,
				"uid" => $map->uid,
				"name" => $map->name,
				"author" => $map->authorNick,
				"file" => $map->fileName
			]);

			//	$sth->execute(['id' => intval($id)]);

			return $sth->fetch();
		}
		catch (\Exception $e)
		{
			echo $e->getMessage();
		}
	}
	private function getPlayerId($login): int
	{
		if ($this->apiClient == null)
			return -1;

		try
		{

			$r = $this->apiClient->sendRequest("playerbylogin/$login", HTTPMethods::GET, []);
			$data = json_decode($r);
			//echo var_export($data, true);
			if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
			{
				return -1;
				echo "YoxCup Api Error : " . $data->{"Message"} . "\n";
			}

			return YoxPlayer::unserialize($data->{"Data"})->id;
		}
		catch (Exception $e)
		{
			return -1;
			echo $e->getMessage() . "\n";
		}
	}
	private function getCup($id)
	{
		$this->logEvent("getCup($id)");
		if ($this->apiClient == null)
		{
			$this->showError("ApiClient unexist ! Try to restart the plugin");
			return;
		}

		$r = $this->apiClient->sendRequest("cup/$id", HTTPMethods::GET, []);
		$data = json_decode($r);

		if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
		{
			$this->showError("YoxCup Api Error : " . $data->{"Message"});
			return false;
		}
		$obj = YoxCup::unserialize($data->{"Values"});
		$this->logEvent(var_export($obj, true));

		if ($obj == null)
		{
			$this->logEvent("getCup == null");
		}

		return $obj;
	}

	/**
	 * @param int $id
	 * @return YoxMatch|false match
	 */
	public function getMatch($id)
	{

		$this->logEvent("getMatch($id)");
		if ($this->curCupID === null || $this->curCupID === 0)
		{
			throw new Exception("L'id de la cup n'est pas défini");
		}
		if ($this->serverId === -1)
		{
			throw new Exception("L'id du serveur n'est pas défini");
		}
		if ($this->apiClient == null)
		{
			$this->showError("ApiClient unexist ! Try to restart the plugin");
			return false;
		}

		$query = "match/" . $this->cup->id . "/$id";
		$r = $this->apiClient->sendRequest($query, HTTPMethods::GET, []);
		$data = json_decode($r);

		if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
		{
			$this->showError("YoxCup Api Error : " . $data->{"Message"});
			return false;
		}

		return YoxMatch::unserialize($data->{"Values"});
	}

	public function getTeam($id)
	{
		$this->logEvent("getTeam($id)");
		if ($this->apiClient == null)
		{
			$this->showError("ApiClient unexist ! Try to restart the plugin");
			return;
		}

		try
		{
			$r = $this->apiClient->sendRequest("team/$id", HTTPMethods::GET, []);
			$data = json_decode($r);

			if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
			{
				$this->showError("YoxCup Api Error : " . $data->{"Message"});
				return false;
			}

			return YoxTeam::unserialize($data->{"Values"});
		}
		catch (\Exception $e)
		{
			echo $e->getMessage() . "\n";
			return false;
		}
	}

	/**
	 * Return a list of players who are in $id team
	 * @return Array
	 */
	public function getTeamPlayers($id)
	{
		$this->logEvent("getTeamPlayers($id)");
		if ($this->apiClient == null)
		{
			$this->showError("ApiClient unexist ! Try to restart the plugin");
			return;
		}

		try
		{
			$r = $this->apiClient->sendRequest("team/$id", HTTPMethods::GET, []);
			$data = json_decode($r);

			if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
			{
				$this->showError("YoxCup Api Error : " . $data->{"Message"});
				return false;
			}

			return YoxPlayer::unserializeArray($data->{"Values"}->{"players"});
		}
		catch (\Exception $e)
		{
			echo $e->getMessage() . "\n";
			return false;
		}
	}

	/**
	 * VÃ©rifie que les joueurs prÃ©sent composent bien les Ã©quipes sÃ©lectionnÃ©es
	 */
	public function checkPlayersTeam()
	{
		$players = $this->maniaControl->getClient()->getPlayerList();
		$ret = true;


		foreach ($players as $player)
		{
			if (!$player->spectator)
			{
				$ret = $this->checkOnePlayerTeam($player->login, $player->teamId);
			}

			if (!$ret)
			{
				break;
			}
		}

		if ($ret)
		{
			$this->maniaControl->getChat()->sendSuccess($this->chatprefix . "Teams are OK !");
		}
		else
		{
			//	$this->maniaControl->getChat()->sendError($this->chatprefix . 'Bad teams !');
		}
	}

	/**
	 * Check in YoxDatabase if a player is in a team
	 * TODO: make API call
	 */
	public function checkOnePlayerTeam($player, $team)
	{
		return (in_array($player, $team->players));
	}

	public function getMapPack($player)
	{
		//TODO: make API call + add check locals maps
		try
		{
			$sth = $this->db->prepare("SELECT yox_maps.uid as uid, yox_maps.name as name, yox_mapmatchs.orderMap as orderMap 
								   FROM yox_mapmatchs 
								   INNER JOIN yox_maps ON yox_mapmatchs.idMap = yox_maps.id
								   WHERE yox_mapmatchs.idMatch = :match 
								   AND yox_mapmatchs.idCup = :cup
								   ORDER BY orderMap");

			$sth->execute([
				"match" => $this->curMatchID,
				"cup" => $this->curCupID
			]);


			$maps = $sth->fetchAll();

			if ($maps === null || count($maps) == 0)
			{
				//	$this->maniaControl->getChat()->sendErrorToAdmins("No mappack for this match");
				return -1;
			}
			else
			{
				$this->matchMapPack = $maps;
				$this->maniaControl->getMapManager()->getMapQueue()->clearMapQueue();


				if ($this->maniaControl->getMapManager()->getCurrentMap()->uid === $maps[0]["uid"])
				{
					if (self::_DEBUG_)
						//	$this->maniaControl->getChat()->sendErrorToAdmins("[DEBUG]>La 1Ã¨re du mappack est la map actuelle");

						for ($i = 1; $i < count($maps); $i++)
						{
							$this->maniaControl->getMapManager()->getMapQueue()->addMapToMapQueue($player->login, $maps[$i]["uid"]);
							$this->maniaControl->getChat()->sendInformation("Map : " . $maps[$i]["name"]);
						}
				}
				else
				{
					foreach ($maps as $map)
					{
						$this->maniaControl->getMapManager()->getMapQueue()->addMapToMapQueue($player->login, $map["uid"]);
						$this->maniaControl->getChat()->sendInformation("Map : " . $map["name"]);
					}
				}
				return count($maps);
			}
		}
		catch (\Exception $e)
		{
			//$this->maniaControl->getChat()->sendErrorToAdmins($e->getMessage());
		}
	}

	private function buildServerList()
	{
		$this->servers = [];
		if ($this->dbConnected && $this->serverId !== -1)
		{
			try
			{
				if ($this->curCupID === -1)
				{
					$r = $this->apiClient->sendRequest("teams", HTTPMethods::GET, []);
					$data = json_decode($r);

					if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
					{
						$this->showError("YoxCup Api Error : " . $data->{"Message"});
						return false;
					}

					$this->servers[] = YoxTeam::unserialize($data->{"Values"});
				}
				else
				{
					$r = $this->apiClient->sendRequest("server/bycup/" . $this->curCupID, HTTPMethods::GET, []);
					$data = json_decode($r);;
					if (isset($data->{"Type"}) && $data->{"Type"} == "Error")
					{
						$this->showError("YoxCup Api Error : " . $data->{"Message"});
						return false;
					}

					$this->servers = YoxTeam::unserializeArray($data->{"Values"});
				}
			}
			catch (\Exception $e)
			{
			}
		}
	}


	//=================================================================================================================================================================
	//==Interface/Manialink============================================================================================================================================
	//=================================================================================================================================================================


	private function buildManialink($login = null)
	{

		if ($this->isLobby)
		{
			$this->buildServerList();

			$labelStyle = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultLabelStyle();
			$quadStyle = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultQuadStyle();
			$quadSubstyle = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultQuadSubstyle();

			$posX = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLSL_POSX);
			$posY = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLSL_POSY);
			$width = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLSL_WIDTH);
			$height = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLSL_HEIGHT);
			$title = "Servers";
			$serverHeight = 8;



			$manialink = new ManiaLink(self::ML_SERVERS_ID);

			$frame = new Frame();
			$manialink->addChild($frame);
			$frame->setPosition($posX, $posY);

			$backgroundQuad = new Quad();
			$frame->addChild($backgroundQuad);
			$backgroundQuad->setSize($width, $height);
			$backgroundQuad->setStyles($quadStyle, $quadSubstyle);

			$titleLabel = new Label();
			$frame->addChild($titleLabel);
			$titleLabel->setScale(0.9);
			$titleLabel->setStyle($labelStyle);
			$titleLabel->setText($title);
			$titleLabel->setTextSize(1);
			$titleLabel->setTranslate(true);
			$titleLabel->setWidth(0.85 * $width);
			$titleLabel->setY(0.45 * $height);

			$posServ = 0.33 * $height;
			foreach ($this->servers as $server)
			{
				if ($server->id != $this->serverId)
				{
					$serverQuad = new Quad();
					$frame->addChild($serverQuad);

					$serverQuad->setSize($width, $serverHeight);
					$serverQuad->setStyles($quadStyle, $quadSubstyle);
					$serverQuad->setY($posServ);

					$titleLabel = new Label();
					$frame->addChild($titleLabel);
					$titleLabel->setScale(0.9);
					$titleLabel->setStyle($labelStyle);
					$titleLabel->setText($server->name);
					$titleLabel->setTextSize(1);
					$titleLabel->setTranslate(true);
					$titleLabel->setWidth(0.85 * $width);
					$titleLabel->setY($posServ - 0.1);
					$titleLabel->setX(-0.3 * $width);

					$playQuad = new Quad_Icons64x64_1();
					$playQuad->setSize(5, 5);
					$playQuad->setSubStyle($playQuad::SUBSTYLE_ArrowBlue);
					$playQuad->setY($posServ - 0.1);
					$playQuad->setX(0.3 * $width);
					$frame->addChild($playQuad);
					//$playQuad->setAction(self::ACT_SERVER_PLAY . "." . $server->getId());
					$playQuad->setUrl("maniaplanet://#join=yoxdeodgeball1v1@GEs@guerro");


					$posServ -= ($serverHeight + 0.1);
				}
			}
			$this->maniaControl->getManialinkManager()->sendManialink($manialink, $login);
		}
		else
		{
			$link = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_GOTO_LOBBY);
			$x = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLGL_POSX);
			$y = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLGL_POSY);
			$sizex = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLGL_WIDTH);
			$sizey = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_MLGL_HEIGHT);
			$maniaLink = new ManiaLink(self::ML_GOTOBUTTON_ID);
			$frame = new Frame();
			$maniaLink->add($frame);
			$frame->setSize($sizex, $sizey);
			$frame->setPosition($x, $y);
			$backgroundQuad = new Quad();
			$frame->add($backgroundQuad);
			$backgroundQuad->setSize($sizex, $sizey);
			$backgroundQuad->setImage("https://upload-icon.s3.us-east-2.amazonaws.com/uploads/icons/png/12883791471558424388-512.png");

			$backgroundQuad->setUrl($link);
			$this->maniaControl->getManialinkManager()->sendManialink($maniaLink, $login);
		}
	}


	public function handleManialinkPageAnswer(array $callback)
	{
		/*$actionId    = $callback[1][2];
		$boolSetting = (strpos($actionId, self::ACT_SERVER_PLAY) === 0);
		if (!$boolSetting)
		{
		return;
		}
		$login       = $callback[1][1];
		$player      = $this->maniaControl->getPlayerManager()->getPlayer($login);
		$actionArray = explode(".", $callback[1][2]);
		$id = intval($actionArray[2]);*/
	}


	//=================================================================================================================================================================
	//==UTIL===========================================================================================================================================================
	//=================================================================================================================================================================

	/**
	 * Check privileges of a player
	 * @param Player $player
	 * @return bool (true if >= Admin)
	 */
	private function checkPrivileges($player)
	{
		$authLevel = "Admin";
		if ($player == null) //Si = null c'est un appel interne
		{
			return true;
		}
		if (!$this->maniaControl->getAuthenticationManager()->checkRight($player, AuthenticationManager::getAuthLevel($authLevel)))
		{
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
			return false;
		}
		if (!$this->dbConnected)
		{
			//	$this->maniaControl->getChat()->sendErrorToAdmins("Not connected to YoxCup ! check credentials");
			return false;
		}

		return true;
	}

	/**
	 * Check if $team contains $login (player)
	 * @param array $team (TODO: create object
	 * @param string $login (player)
	 */
	public function util_teamHasPlayer($team, $login)
	{
		foreach ($team as $player)
		{
			echo $player["login"] . "\n";
			if ($player["login"] == $login)
				return true;
		}
		return false;
	}

	private function showError($message, $login = null)
	{
		return;
		if ($login == null)
			$this->maniaControl->getChat()->sendErrorToAdmins("YoxCup>" . $message, AuthenticationManager::AUTH_LEVEL_ADMIN);
		else
			$this->maniaControl->getChat()->sendError("YoxCup>" . $message, $login);

		if (self::_DEBUG_)
		{
			echo "YoxCup>$message\n";
		}
	}




	/**
	 * @see \ManiaControl\Plugins\Plugin::unload()
	 */
	public function unload()
	{
		$this->maniaControl = null;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getId()
	 */
	public static function getId()
	{
		return self::ID;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getName()
	 */
	public static function getName()
	{
		return self::NAME;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getVersion()
	 */
	public static function getVersion()
	{
		return self::VERSION;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getAuthor()
	 */
	public static function getAuthor()
	{
		return self::AUTHOR;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getDescription()
	 */
	public static function getDescription()
	{
		return 'YoxCup Match Manager';
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::prepare()
	 */
	public static function prepare(ManiaControl $maniaControl)
	{
		// Implement prepare() method.
	}

	//Subclasses ======================



}
